# -*- coding: utf-8 -*-

import hmac
import hashlib

from .compact import b, s
from .util import standard_b64encode, _check_dict
from .logger import logger
from functools import reduce


class Auth(object):
    """
    UCloud API 认证类
    对UCloud UFile 上传认证、下载签名和API调用签名接口的具体实现
    """

    def __init__(self, public_key, private_key):
        """
        初始化 Auth 实例

        :param public_key: string类型, 账户公私钥中的公钥
        :param private_key: string类型，账户公私钥中的私钥
        :return 无，当采用非法的公私钥会抛出异常
        """

        self.__checkkey(public_key, private_key)
        self.__public_key = public_key
        self.__private_key = private_key

    def set_keys(self, public_key, private_key):
        """
        重新设置账户API的公私钥

        :param public_key:  string类型, 账户公私钥中的公钥
        :param private_key: string类型, 账户公私钥中的私钥
        :return: 无，当采用非法的公私钥会抛出异常
        """

        self.__checkkey(public_key, private_key)
        self.__public_key = public_key
        self.__private_key = b(private_key)

    def bucket_signature(self, query):
        """
        UCloud UFile 空间管理签名

        :param query: 一个键值对类型为'string: string'字典类型的HTTP请求参数，比如{'Action': 'CreateBucket', 'BucketName': 'bucket'}
        :return: string 类型，UCloud UFile空间管理签名
        """

        _check_dict(query)
        query['PublicKey'] = self.__public_key
        sign_string = reduce(lambda y, x: ''.join([x, query[x], y]), sorted(query.keys(), reverse=True), self.__private_key)
        hashstr = hashlib.sha1(b(sign_string))
        return s(hashstr.hexdigest())

    def ufile_signature(self, data):
        """
        UCloud UFile 下载签名

        :param data: String 类型，待授权签名的字符串
        :return: string类型，采用standard_b64encode编码的UCloud UFile下载签名字符串
        """

        hashstr = hmac.new(b(self.__private_key), b(data), hashlib.sha1)
        return standard_b64encode(hashstr.digest())

    def ufile_authorization(self, data):
        """
        UCloud UFile 上传凭证

        :param data: string 类型，待计算上传凭证的字符串
        :return: string类型，UCloud UFile上传凭证的字符串
        """

        signature = self.ufile_signature(data)
        return '{0} {1}:{2}'.format('UCloud', self.__public_key, signature)

    def _public_key(self):
        """
        获得UCloud 账户API公钥

        :return: string类型，UCloud账户API中的公钥
        """

        return self.__public_key

    @staticmethod
    def __checkkey(public_key, private_key):
        if not (public_key and private_key) or not isinstance(public_key, str) or not isinstance(private_key, str):
            raise ValueError('invalid API keys')


# -*- coding: utf-8 -*-

import re

from requests import Session, RequestException
from requests.adapters import HTTPAdapter

from . import config
from .logger import logger

requests = Session()
requests.mount("http://",
               HTTPAdapter(pool_connections=config.CONNECTION_POOL_SIZE, pool_maxsize=config.CONNECTION_POOL_SIZE))
requests.mount("https://",
               HTTPAdapter(pool_connections=config.CONNECTION_POOL_SIZE, pool_maxsize=config.CONNECTION_POOL_SIZE))


def __return_wraper(response, content_consumed=False):
    """
    UCloud UFile 服务器响应的封装

    :param response: requests response object
    :param content_consumed: boolean类型, 如果响应内容被保存在文件中则为真
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    if response.status_code not in [200, 204, 206]:
        return None, ResponseInfo(response)
    content_type = response.headers.get('Content-Type')
    ret = {} if content_consumed else response.json() if isinstance(content_type, str) and content_type.startswith(
        'application/json') and response.text != "" else {}
    return ret, ResponseInfo(response, None, content_consumed)


def _put_file(url, header, uploadfile):
    """
    采用普通put方法上传文件到UFile空间

    :param url: string类型，上传的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :param uploadfile: 本地文件名称
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    with open(uploadfile, 'rb') as data:
        return _put_stream(url, header, data)


def _put_stream(url, header, data):
    """
    采用普通方法上传二进制流到UFile空间

    :param url: string类型，上传的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :param data: 二进制数据流
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.put(url, headers=header, data=data, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _post_file(url, header, data):
    """
    采用表单上传的方法上传文件到空间

    :param url:  string类型，上传的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :param data: 二进制数据流
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """

    try:
        response = requests.post(url, headers=header, data=data, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _uploadhit_file(url, header, params):
    """
    秒传文件到空间

    :param url:string类型， 秒传文件的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :param params: dict类型，http 请求的查询参数，键值对类型分别为string类型
    :return: ret: return message, None if response status code not in [200, 204, 206] else a dict-like object with response body
    :return: ResponseInfo: UCloud UFile server response info
    """

    try:
        response = requests.post(url, params=params, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _delete_file(url, header):
    """
    删除文件

    :param url: string类型, 要删除文件的url
    :param header: dict 类型，键值对类型分别为string类型，HTTP请求头
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.delete(url, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _head_file(url, header):
    """
    获取文件信息

    :param url: string类型, 要删除文件的url
    :param header: dict 类型，键值对类型分别为string类型，HTTP请求头
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.head(url, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _initialsharding(url, header):
    """
    初始化分片请求

    :param url:  string类型, 初始化分片请求的url
    :param header: dict 类型，键值对类型分别为string类型，HTTP请求头
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.post(url, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _shardingupload(url, data, header):
    """
    分片上传数据

    :param url: string类型，分片上传数据的url
    :param data: bytes 类型，分片上传的二进制数据
    :param header: dict 类型，键值对类型分别为string类型，HTTP请求头
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.put(url, headers=header, data=data, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _finishsharding(url, param, header, data):
    """
    结束分片上传请求

    :param url:  string类型, 初始化分片请求的url
    :param header: dict 类型，键值对类型分别为string类型，HTTP请求头
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.post(url, headers=header, params=param, data=data,
                                 timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _download_file(url, header, localfile):
    """
    下载文件

    :param url: string类型, 下载UFile资源的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :param localfile: string类型, 保存文件的本地名称
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.get(url, headers=header, stream=True)
    except RequestException as e:
        return None, ResponseInfo(None, e)
    if response.status_code in [200, 206]:
        with open(localfile, 'wb') as fd:
            for block in response.iter_content(config.BLOCKSIZE):
                fd.write(block)
    else:
        return __return_wraper(response)
    return __return_wraper(response, True)


def _getfilelist(url, header, param):
    """
    获取文件列表

    :param url: string 类型，获取文件列表的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """

    try:
        response = requests.get(url, headers=header, params=param, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        logger.error('send request error:{0}'.format(e))
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _listobjects(url, header, param):
    """
    获取目录文件列表

    :param url: string 类型，获取目录文件列表的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """

    try:
        response = requests.get(url, headers=header, params=param, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        logger.error('send request error:{0}'.format(e))
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _restore_file(url, header):
    """
    解冻冷存文件请求

    :param url:  string类型, 解冻请求的url
    :param header: dict 类型，键值对类型分别为string类型，HTTP请求头
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """

    try:
        response = requests.put(url, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _classswitch_file(url, header, params):
    """
    文件存储类型转换请求

    :param url:string类型，文件存储类型转换的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :param params: dict类型，http 请求的查询参数，键值对类型分别为string类型
    :return: ret: return message, None if response status code not in [200, 204, 206] else a dict-like object with response body
    :return: ResponseInfo: UCloud UFile server response info
    """

    try:
        response = requests.put(url, params=params, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _copy_file(url, header):
    """
    拷贝文件到空间

    :param url:string类型， 拷贝文件的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :return: ret: return message, None if response status code not in [200, 204, 206] else a dict-like object with response body
    :return: ResponseInfo: UCloud UFile server response info
    """

    try:
        response = requests.put(url, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _rename_file(url, header, params):
    """
    重命名文件到空间

    :param url:string类型， 重命名文件的url
    :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
    :param params: dict类型，http 请求的查询参数，键值对类型分别为string类型
    :return: ret: return message, None if response status code not in [200, 204, 206] else a dict-like object with response body
    :return: ResponseInfo: UCloud UFile server response info
    """

    try:
        response = requests.put(url, params=params, headers=header, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _bucket_request(url, param, header):
    """
    UCloud UFile 空间管理请求

    :param url: String类型，空间管理请求的url
    :param param: dict类型，键值对类型分别为string类型，HTTP请求的查询参数
    :param header: dict类型，键值对类型分别为string类型，HTTP请求头信息
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.get(url, headers=header, params=param)
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _get_multi_upload_part(url, header):
    """
    UCloud UFile 获取分片上传列表信息请求

    :param url: String类型，获取分片上传列表信息请求的url
    :param header: dict类型，键值对类型分别为string类型，HTTP请求头信息
    :return: ret: 如果http状态码不为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
    :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
    """
    try:
        response = requests.get(url, headers=header)
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


def _op_meta(url, body, header):
    """
    UCloud UFile 文件元数据操作请求

    :param url: String类型，文件元数据操作请求的url
    :param body: dict类型，键值对类型分别为string类型，HTTP请求的body信息
    :param header: dict类型，键值对类型分别为string类型，HTTP请求头信息
    """
    try:
        response = requests.post(url, headers=header, data=body, timeout=config.get_default('connection_timeout'))
    except RequestException as e:
        return None, ResponseInfo(None, e)
    return __return_wraper(response)


class ResponseInfo(object):
    """
    UCloud UFile 服务器返回信息,解析UCloud UFile服务器返回信息以及网络连接问题

    Attributes:
        content: 服务器返回信息的二级制数据块，如果出现网络连接问题则为None
        status_code: Integer类型，服务器返回信息状态码，如果出现网络连接问题则为-1
        error:       string类型，服务器返回错误信息或者网络连接问题
        x_session_id: String类型，UFile服务器处理请求失败响应的会话ID，如果正常或出现网络连接问题则为None
        content_type: string，UFile服务器返回信息的类型，如果出现网络连接问题则为None
        content_length: integer类型，服务器返回信息长度，如果出现网络连接问题则为None
        content_range: 文件下载的数据范围，None或者元素为两个整数的元组
        ret_code:       Integer类型, UCloud UFile 服务内部错误码,正确服务或者网络连接则为0
        etag:           string类型,文件的etag，无则为None
    """

    def __init__(self, response, exception=None, content_consumed=False):
        """
        初始化 ResponseInfo 实例

        :param response: requests response object
        :param exception: exception object
        :param content_consumed: boolean 类型, 如果响应内容被保存则为True
        :return: None
        """
        self.__response = response
        self.exception = exception

        if response is None:
            self.status_code = -1
            self.error = str(exception)
            self.x_session_id = None
            self.etag = None
            self.content_type = None
            self.content_length = None
            self.content_range = None
            self.ret_code = None
            self.content = None
            self.md5 = None
            self.x_ufile_restore = None
        else:
            self.status_code = response.status_code
            self.x_session_id = response.headers.get('X-SessionId')
            self.content_type = response.headers.get('Content-Type')
            self.md5 = response.headers.get('Content-MD5')
            self.content_length = response.headers.get('Content-Length')
            content_length = response.headers.get('content-range')
            self.content = None if content_consumed else response.content
            self.x_ufile_restore = response.headers.get('X-Ufile-restore')
            if content_length is not None:
                byteslist = re.split('[- /]', content_length)
                self.content_range = (int(byteslist[1]), int(byteslist[2]))
            else:
                self.content_range = None
            self.etag = response.headers.get('Etag')
            if self.status_code not in [200, 204, 206]:
                ret = response.json() if response.headers.get(
                    'Content-Type') == 'application/json' and response.text != "" else None
                if ret is None:
                    self.error = 'unknown error'
                    self.ret_code = -1
                else:
                    self.ret_code = ret['RetCode']
                    self.error = ret['ErrMsg']
            else:
                self.error = None
                self.ret_code = None

    def ok(self):
        """
        返回请求是否成功

        :return: boolean类型, 请求成功则返回True
        """
        return self.status_code in [200, 204, 206]

    def need_retry(self):
        """
        返回是否由于网络连接异常需要重新请求

        :return: boolean 类型
        """
        return self.__response is None

    def __str__(self):
        """

        :return: string类型，Response的信息
        """

        return ', '.join(['{0}: {1}'.format(key, value) for (key, value) in self.__dict__.items() if key != 'content'])

    def __repr__(self):
        """
        The response info

        :return:: string
        """
        return self.__str__



# -*- coding: utf-8 -*-

import json
import math
import os
import threading
import time

from . import config
from .baseufile import BaseUFile
from .compact import s
from .httprequest import ResponseInfo, _initialsharding, _finishsharding, _shardingupload, _get_multi_upload_part
from .logger import logger
from .util import _check_dict, initialsharding_url, finishsharding_url, shardingupload_url, _file_iter, \
    mimetype_from_file, mimetype_from_buffer, deprecated, ufile_listparts_url


class MultipartUploadUFile(BaseUFile):
    """
    分片上传UFile类

    Attributes:
    uploadid: string类型，初始化分片上传请求获得的uploadid
    etaglist: list 类型，已经成功上传分片的etag列表
    blocksize: integer类型，分片大小
    pausepartnumber: 第一个上传失败的分片编号（编号从0开始）
    """

    def __init__(self, public_key, private_key, upload_suffix=None):
        """
        初始化 MultipartUploadUFile 实例

        :param public_key: string类型, 账户API公私钥中的公钥
        :param private_key: string类型, 账户API公私钥中的私钥
        :param upload_suffix: string类型, 如果传入此参数, 则会忽略 config 中配置的 upload_suffix 字段
        :return: None，如果为非法的公私钥，则抛出ValueError异常
        """
        super(MultipartUploadUFile, self).__init__(public_key, private_key)
        self.uploadid = None
        self.blocksize = None
        self.etaglist = []
        self.__key = None
        self.__bucket = None
        self.__header = None
        self.__datastream = None
        self.__mimetype = None
        self.__upload_suffix = upload_suffix
        self.pausepartnumber = 0

    def uploadstream(self, bucket, key, stream, maxthread=4, retrycount=3, retryinterval=5, mime_type=None,
                     header=None):
        """
        分片上传二进制数据流到UFile空间

        :param bucket: 空间名称
        :param key: 上传数据在空间中的名称
        :param stream: file-like 对象或者二进制数据流
        :param maxthread: 限制的最大线程数
        :param mime_type: 上传数据的MIME类型
        :param retrycount: integer 类型，分片重传次数
        :param retryinterval: integer 类型，同个分片失败重传间隔，单位秒
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        self.__bucket = bucket
        self.__key = key
        self.etaglist = []
        self.uploadid = None
        self.blocksize = None
        self.__header = header
        self.__stream = stream
        self.pausepartnumber = 0

        self.__threaddict = {}  # 线程字典，用于保证finish前等待运行中的uploadpart线程
        self.__errresp = None  # 用于主函数返回的errresp

        if self.__header is None:
            self.__header = dict()
        else:
            _check_dict(self.__header)
        if 'User-Agent' not in self.__header:
            self.__header['User-Agent'] = config.get_default('user_agent')

        # initial sharding request
        ret, resp = self.__initialsharding()
        if resp.ok():
            self.uploadid = ret.get('UploadId')
            self.blocksize = ret.get('BlkSize')
            logger.info('multipart upload id: {0}'.format(self.uploadid))
        else:
            logger.error('multipart upload init failed. error message: {0}'.format(resp.error))
            return ret, resp

        # mulitple sharding upload
        if mime_type is None:
            if hasattr(self.__stream, 'seek') and hasattr(self.__stream, 'read'):
                self.__mimetype = s(mimetype_from_buffer(self.__stream.read(1024)))
                self.__stream.seek(0, os.SEEK_SET)
            else:
                self.__mimetype = 'application/octec-stream'
        else:
            self.__mimetype = mime_type

        self.__header['Content-Type'] = self.__mimetype
        authorization = self.authorization('put', self.__bucket, self.__key, self.__header)
        self.__header['Authorization'] = authorization

        sem = threading.Semaphore(maxthread)
        partnumber = 0
        for data in _file_iter(self.__stream, self.blocksize):
            sem.acquire()  # 控制最大并发线程数
            if self.__errresp:  # 如果有分片上传失败，停止后续分片上传
                sem.release()
                break
            self.etaglist.append("")
            thread1 = threading.Thread(target=self.__partthread, args=(
                sem, self.__bucket, self.__key, self.uploadid, partnumber, self.__header, data, retrycount,
                retryinterval,
                self.etaglist, self.__upload_suffix))
            self.__threaddict[partnumber] = thread1
            thread1.start()
            partnumber += 1

        for thread in list(self.__threaddict.values()):  # 转为list是因为遍历字典时长度变化会报错
            thread.join()

        if self.__errresp:
            self.pausepartnumber = self.etaglist.index("")
            logger.error(
                'multipart upload failed. uploadid:{0}, pausepartnumber: {1}, key: {2} , error message{3}. FAIL!!!'.format(
                    self.uploadid, self.pausepartnumber, self.__key, self.__errresp.error))
            return None, self.__errresp
        else:
            self.pausepartnumber = len(self.etaglist)

        logger.info('start finish sharding request.')
        ret, resp = self.__finishupload()
        if not resp.ok():
            logger.error(
                'multipart upload failed. uploadid:{0}, pausepartnumber: {1}, key: {2}, error message: {3}. FAIL!!!'.format(
                    self.uploadid, self.pausepartnumber, self.__key, resp.error))
        else:
            logger.info('mulitpart upload succeed. uploadid: {0}, key: {1} SUCCEED'.format(self.uploadid, self.__key))
        return ret, resp

    def uploadfile(self, bucket, key, localfile, maxthread=4, retrycount=3, retryinterval=5, header=None):
        """
        分片上传本地文件到空间

        :param bucket: string类型，空间名称
        :param key: string类型，文件在空间中的名称
        :param localfile: string 类型，本地文件名称
        :param maxthread: 限制的最大线程数
        :param retrycount: integer 类型，分片重传次数
        :param retryinterval: integer 类型，同个分片失败重传间隔，单位秒
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        self.__localfile = localfile
        mime_type = s(mimetype_from_file(self.__localfile))
        with open(localfile, 'rb') as fd:
            return self.uploadstream(bucket, key, fd, maxthread, retrycount, retryinterval, mime_type, header)

    def __initialsharding(self):
        """
        初始化分片请求

        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if self.__header is None:
            self.__header = {}
        else:
            _check_dict(self.__header)
        if 'User-Agent' not in self.__header:
            self.__header['User-Agent'] = config.get_default('user_agent')

        self.__header['Content-Length'] = str(0)
        self.__header['Content-Type'] = 'text/plain'
        authorization = self.authorization('post', self.__bucket, self.__key, self.__header)
        self.__header['Authorization'] = authorization

        url = initialsharding_url(self.__bucket, self.__key, upload_suffix=self.__upload_suffix)

        logger.info('start initialize sharding')
        logger.info('initial sharding url: {0}'.format(url))

        return _initialsharding(url, self.__header)

    def __finishupload(self):
        """
        完成分片上传的请求

        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if self.__header is None:
            self.__header = dict()
        else:
            _check_dict(self.__header)
        if 'User-Agent' not in self.__header:
            self.__header['User-Agent'] = config.get_default('user_agent')

        self.__header['Content-Type'] = 'text/plain'
        logger.info(self.etaglist)
        data = ','.join(self.etaglist)
        logger.info(data)
        self.__header['Content-Length'] = str(len(data))

        authorization = self.authorization('post', self.__bucket, self.__key, self.__header)
        self.__header['Authorization'] = authorization
        logger.info(json.dumps(self.__header, indent=4))
        params = {'uploadId': self.uploadid}

        url = finishsharding_url(self.__bucket, self.__key, upload_suffix=self.__upload_suffix)

        logger.info('start finish sharding request')
        return _finishsharding(url, params, self.__header, data)

    @deprecated("Deprecated since version 3.2.6")
    def resumeuploadfile(self, retrycount=3, retryinterval=5, bucket=None, key=None, uploadid=None, blocksize=None,
                         etaglist=None, localfile=None, pausepartnumber=None, mime_type=None, header=None):
        """
        断点续传失败的本地文件分片
        可以在调用uploadfile失败后重新续传，也可以通过传递所有需要的参数续传

        :param retrycount: integer 类型，分片重传次数
        :param retryinterval: integer 类型，同个分片失败重传间隔，单位秒
        :param bucket: string类型, 空间名称
        :param key: string类型，文件或数据在空间中的名称
        :param uploadid: string类型，初始化分片获得的uploadid
        :param blocksize: integer类型，分片大小
        :param etaglist: list类型，元素为已经上传成功的分片的etag
        :param pausepartnumber: integer类型，第一个失败分片的编号（编号从0开始）
        :param localfile: string类型，本地文件名称
        :param mime_type: string类型，上传数据的MIME类型
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if localfile:
            self.__localfile = localfile
        if blocksize:
            self.blocksize = blocksize
        if pausepartnumber:
            self.pausepartnumber = pausepartnumber
        if mime_type:
            self.__mimetype = mime_type
        else:
            mime_type = s(mimetype_from_file(self.__localfile))
        with open(self.__localfile, 'rb') as fd:
            fd.seek(self.pausepartnumber * self.blocksize, os.SEEK_SET)
            return self.resumeuploadstream(retrycount, retryinterval, bucket, key, uploadid, blocksize, etaglist, fd,
                                           pausepartnumber, mime_type, header)

    def resumeuploadstream(self, retrycount=3, retryinterval=5, bucket=None, key=None, uploadid=None, blocksize=None,
                           etaglist=None, stream=None, pausepartnumber=None, mime_type=None, header=None):
        """
        断点续传失败数据流的分片
        可以在调用uploadstream失败后重新续传，也可以通过传递所有需要的参数续传

        :param retrycount: integer 类型，分片重传次数
        :param retryinterval: integer 类型，同个分片失败重传间隔，单位秒
        :param bucket: string类型, 空间名称
        :param key: string类型，文件或数据在空间中的名称
        :param uploadid: string类型，初始化分片获得的uploadid
        :param blocksize: integer类型，分片大小
        :param etaglist: list类型，元素为已经上传成功的分片的etag
        :param pausepartnumber: integer类型，第一个失败分片的编号（编号从0开始）
        :param stream: file-like对象或者二进制数据流，需要重新上传的数据
        :param mime_type: string类型，上传数据的MIME类型
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """

        if bucket:
            self.__bucket = bucket
        if key:
            self.__key = key
        if uploadid:
            self.uploadid = uploadid
        if blocksize:
            self.blocksize = blocksize
        if stream:
            self.__stream = stream
        if etaglist:
            self.etaglist = etaglist
        if pausepartnumber:
            self.pausepartnumber = pausepartnumber
        if header:
            self.__header = header
        if mime_type is not None:
            self.__mimetype = mime_type
        elif self.__mimetype is None:
            self.__mimetype = 'application/octec-stream'

        if self.__header is None:
            self.__header = dict()
        else:
            _check_dict(self.__header)
        if 'User-Agent' not in self.__header:
            self.__header['User-Agent'] = config.get_default('user_agent')

        # initial sharding request
        if self.uploadid is None:
            ret, resp = self.__initialsharding()
            if resp.ok():
                self.uploadid = ret.get('UploadId')
                self.blocksize = ret.get('BlkSize')
                logger.info('multipart upload id: {0}'.format(self.uploadid))
            else:
                logger.error('multipar upload init failed. error message: {0}'.format(resp.error))
                return ret, resp

        self.__header['Content-Type'] = self.__mimetype
        authorization = self.authorization('put', self.__bucket, self.__key, self.__header)
        self.__header['Authorization'] = authorization

        for data in _file_iter(self.__stream, self.blocksize):
            url = shardingupload_url(self.__bucket, self.__key, self.uploadid, self.pausepartnumber)
            ret = None
            resp = None
            for index in range(retrycount):
                logger.info('retry {0} time sharding upload sharing {1}'.format(index + 1, self.pausepartnumber))
                logger.info('sharding url:{0}'.format(url))
                ret, resp = _shardingupload(url, data, self.__header)
                if not resp.ok():
                    logger.error(
                        'failed {0} time when retry upload sharding {1},error message: {2}, uploadid: {3}'.format(
                            index + 1, self.pausepartnumber, resp.error, self.uploadid))
                    if index < retrycount - 1:
                        time.sleep(retryinterval)
                else:
                    break
            if not resp.ok():
                logger.error(
                    'retry upload sharding {0} failed, uploadid: {1}'.format(self.pausepartnumber, self.uploadid))
                return ret, resp
            logger.info(
                'retry upload sharding {0} succeed. etag: {1}, uploadid: {2}'.format(self.pausepartnumber, resp.etag,
                                                                                     self.uploadid))
            self.pausepartnumber += 1
            self.etaglist.append(resp.etag)
        # finish sharding upload
        logger.info('start finish upload request')
        ret, resp = self.__finishupload()
        if not resp.ok():
            logger.error(
                'multipart upload failed. uploadid:{0}, pausepartnumber: {1}, key: {2} FAIL!!!'.format(self.uploadid,
                                                                                                       self.pausepartnumber,
                                                                                                       self.__key))
        else:
            logger.info(
                'mulitpart upload succeed. uploadid: {0}, key: {1} SUCCEED!!!'.format(self.uploadid, self.__key))
        return ret, resp

    def __partthread(self, sem, bucket, key, uploadid, part_number, header, data, retrycount, retryinterval, etaglist,
                     upload_suffix=None):
        url = shardingupload_url(bucket, key, uploadid, part_number, upload_suffix=upload_suffix)
        resp = None
        for index in range(retrycount):
            logger.info('try {0} time sharding upload sharding {1}'.format(index + 1, part_number))
            logger.info('sharding url:{0}'.format(url))
            _, resp = _shardingupload(url, data, header)
            if not resp.ok():
                logger.error(
                    'failed {0} time when upload sharding {1}.error message: {2}, uploadid: {3}'.format(index + 1,
                                                                                                        part_number,
                                                                                                        resp.error,
                                                                                                        uploadid))
                if index < retrycount - 1:
                    time.sleep(retryinterval)
            else:
                break

        if not resp.ok():
            logger.error('upload sharding {0} failed. uploadid: {1}'.format(part_number, uploadid))
            self.__errresp = resp
        else:
            logger.info('upload sharding {0} succeed.etag:{1}, uploadid: {2}'.format(part_number, resp.etag, uploadid))
            etaglist[part_number] = resp.etag
        self.__threaddict.pop(part_number)
        sem.release()

    def init_multipart_upload(self, bucket, key, header=None, upload_suffix=None):
        """
        初始化分片请求

        :param bucket: string类型，空间名称
        :param key: string类型，文件或数据在空间中的名称
        :param header: dict类型，HTTP请求头部
        :param upload_suffix: string类型, 如果传入此参数, 则会忽略 config 中配置的 upload_suffix 字段
        :return: (dict, ResponseInfo) tuple类型，dict为返回的json信息，ResponseInfo为请求的返回信息
        """
        self.__bucket = bucket
        self.__key = key
        self.__header = header if header else dict()
        if 'User-Agent' not in self.__header:
            self.__header['User-Agent'] = config.get_default('user_agent')
        self.__upload_suffix = upload_suffix if upload_suffix else config.get_default('upload_suffix')
        ret, resp = self.__initialsharding()
        if resp.ok:
            self.uploadid = ret.get('UploadId')
            self.blocksize = ret.get('BlkSize')
        return ret, resp

    def finish_upload(self):
        """
        完成分片请求

        :return: (dict, ResponseInfo) tuple类型，dict为返回的json信息，ResponseInfo为请求的返回信息
        """
        return self.__finishupload()

    def upload_part_copy(self, source_bucket, source_key, mime_type, offset, size, part_number, upload_suffix=None):
        """
        分片COPY上传

        :param source_bucket: string类型, 空间名称
        :param source_key: string类型，文件或数据在空间中的名称
        :param mime_type: string类型，上传数据的MIME类型
        :param offset: int类型, copy文件的起始位置
        :param size: int类型, copy文件的大小, 不能大于源文件大小
        :param part_number: int类型, 分片号
        :param upload_suffix: string类型, 如果传入此参数, 则会忽略 config 中配置的 upload_suffix 字段
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        self.__mimetype = mime_type
        self.__upload_suffix = config.get_default("upload_suffix") if upload_suffix is None else upload_suffix
        self.__header.update({"X-Ufile-Copy-Source": "/%s/%s" % (source_bucket, source_key)})
        self.__header.update({"Content-Type": self.__mimetype})
        self.__header.update({"X-Ufile-Copy-Source-Range": "bytes=%d-%d" % (offset, size - 1)})
        if 'User-Agent' not in self.__header:
            self.__header['User-Agent'] = config.get_default('user_agent')
        authorization = self.authorization('put', self.__bucket, self.__key, self.__header)
        self.__header.update({"Authorization": authorization})

        url = shardingupload_url(self.__bucket, self.__key, self.uploadid, part_number,
                                 upload_suffix=self.__upload_suffix)
        ret, resp = _shardingupload(url, None, self.__header)
        if resp.ok:
            self.etaglist.append(resp.etag)
        return ret, resp

    def get_multi_upload_part(self, bucket, upload_id, max_parts=None, part_number_marker=None, header=None, upload_suffix=None):
        """
        获取未完成分片上传的对象的已上传成功的分片列表。

        :param bucket: string类型，空间名称
        :param upload_id: string类型，初始化分片上传时返回的uploadid
        :param max_parts: int类型，返回的最大条目数，默认为1000
        :param part_number_marker: int类型，返回的起始条目，默认为0
        :param header: dict类型，HTTP请求头部
        :param upload_suffix: string类型, 如果传入此参数, 则会忽略 config 中配置的 upload_suffix 字段
        :return: (dict, ResponseInfo) tuple类型，dict为返回的json信息，ResponseInfo为请求的返回信息
        """
        self.__header = header if header else dict()
        if 'User-Agent' not in self.__header:
            self.__header['User-Agent'] = config.get_default('user_agent')
        if upload_suffix is not None:
            self.__upload_suffix = upload_suffix
        authorization = self.authorization('get', bucket, "", self.__header, action='muploadpart')
        self.__header.update({"Authorization": authorization})
        url = ufile_listparts_url(bucket, self.__upload_suffix, upload_id, max_parts, part_number_marker)
        return _get_multi_upload_part(url, self.__header)



# -*- coding: utf-8 -*-

import base64
import hashlib
import mimetypes
import os
import struct
import warnings
from os import path

from . import config
from .compact import *
from .config import BLOCKSIZE, get_default

_EXTRA_TYPES_MAP = {
    ".js": "application/javascript",
    ".xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ".xltx": "application/vnd.openxmlformats-officedocument.spreadsheetml.template",
    ".potx": "application/vnd.openxmlformats-officedocument.presentationml.template",
    ".ppsx": "application/vnd.openxmlformats-officedocument.presentationml.slideshow",
    ".pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
    ".sldx": "application/vnd.openxmlformats-officedocument.presentationml.slide",
    ".docx": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    ".dotx": "application/vnd.openxmlformats-officedocument.wordprocessingml.template",
    ".xlam": "application/vnd.ms-excel.addin.macroEnabled.12",
    ".xlsb": "application/vnd.ms-excel.sheet.binary.macroEnabled.12",
    ".apk": "application/vnd.android.package-archive",
    ".ipa": "application/vnd.ios.package-archive",
    ".323": "text/h323",
    ".3gp": "video/3gpp",
    ".3gpp": "video/3gpp",
    ".7z": "application/x-7z-compressed",
    ".acx": "application/internet-property-stream",
    ".ai": "application/postscript",
    ".aif": "audio/x-aiff",
    ".aifc": "audio/x-aiff",
    ".aiff": "audio/x-aiff",
    ".asf": "video/x-ms-asf",
    ".asr": "video/x-ms-asf",
    ".asx": "video/x-ms-asf",
    ".atom": "application/atom+xml",
    ".au": "audio/basic",
    ".avi": "video/x-msvideo",
    ".axs": "application/olescript",
    ".bas": "text/plain",
    ".bcpio": "application/x-bcpio",
    ".bin": "application/octet-stream",
    ".bmp": "image/bmp",
    ".c": "text/plain",
    ".cat": "application/vnd.ms-pkiseccat",
    ".cco": "application/x-cocoa",
    ".cdf": "application/x-cdf",
    ".cer": "application/x-x509-ca-cert",
    ".class": "application/octet-stream",
    ".clp": "application/x-msclip",
    ".cmx": "image/x-cmx",
    ".cod": "image/cis-cod",
    ".cpio": "application/x-cpio",
    ".crd": "application/x-mscardfile",
    ".crl": "application/pkix-crl",
    ".crt": "application/x-x509-ca-cert",
    ".csh": "application/x-csh",
    ".css": "text/css",
    ".dcr": "application/x-director",
    ".der": "application/x-x509-ca-cert",
    ".dir": "application/x-director",
    ".dll": "application/octet-stream",
    ".dms": "application/octet-stream",
    ".doc": "application/msword",
    ".docx": "application/vnd.openxmlformats-officedocument.wordprocessingml.document",
    ".dot": "application/msword",
    ".dvi": "application/x-dvi",
    ".dxr": "application/x-director",
    ".ear": "application/java-archive",
    ".eot": "application/vnd.ms-fontobject",
    ".eps": "application/postscript",
    ".etx": "text/x-setext",
    ".evy": "application/envoy",
    ".exe": "application/octet-stream",
    ".fif": "application/fractals",
    ".flr": "x-world/x-vrml",
    ".flv": "video/x-flv",
    ".gif": "image/gif",
    ".gtar": "application/x-gtar",
    ".gz": "application/x-gzip",
    ".h": "text/plain",
    ".hdf": "application/x-hdf",
    ".hlp": "application/winhlp",
    ".hqx": "application/mac-binhex40",
    ".hta": "application/hta",
    ".htc": "text/x-component",
    ".htm": "text/html",
    ".html": "text/html",
    ".htt": "text/webviewhtml",
    ".ico": "image/x-icon",
    ".ief": "image/ief",
    ".iii": "application/x-iphone",
    ".ins": "application/x-internet-signup",
    ".isp": "application/x-internet-signup",
    ".jad": "text/vnd.sun.j2me.app-descripto",
    ".jar": "application/java-archive",
    ".jardiff": "application/x-java-archive-diff",
    ".jfif": "image/pipeg",
    ".jng": "image/x-jng",
    ".jnlp": "application/x-java-jnlp-file",
    ".jpe": "image/jpeg",
    ".jpeg": "image/jpeg",
    ".jpg": "image/jpeg",
    ".js": "application/x-javascript",
    ".json": "application/json",
    ".kar": "audio/midi",
    ".kml": "application/vnd.google-earth.kml+xml",
    ".kmz": "application/vnd.google-earth.kmz",
    ".latex": "application/x-latex",
    ".lha": "application/octet-stream",
    ".lsf": "video/x-la-asf",
    ".lsx": "video/x-la-asf",
    ".lzh": "application/octet-stream",
    ".m13": "application/x-msmediaview",
    ".m14": "application/x-msmediaview",
    ".m3u": "audio/x-mpegurl",
    ".m3u8": "application/vnd.apple.mpegurl",
    ".m4a": "audio/x-m4a",
    ".m4v": "video/x-m4v",
    ".man": "application/x-troff-man",
    ".mdb": "application/x-msaccess",
    ".me": "application/x-troff-me",
    ".mht": "message/rfc822",
    ".mhtml": "message/rfc822",
    ".mid": "audio/midi",
    ".midi": "audio/midi",
    ".mml": "text/mathml",
    ".mng": "video/x-mng",
    ".mny": "application/x-msmoney",
    ".mov": "video/quicktime",
    ".movie": "video/x-sgi-movie",
    ".mp2": "video/mpeg",
    ".mp3": "audio/mpeg",
    ".mp4": "video/mp4",
    ".mpa": "video/mpeg",
    ".mpe": "video/mpeg",
    ".mpeg": "video/mpeg",
    ".mpg": "video/mpeg",
    ".mpp": "application/vnd.ms-project",
    ".mpv2": "video/mpeg",
    ".ms": "application/x-troff-ms",
    ".mvb": "application/x-msmediaview",
    ".nws": "message/rfc822",
    ".oda": "application/oda",
    ".ogg": "audio/ogg",
    ".p10": "application/pkcs10",
    ".p12": "application/x-pkcs12",
    ".p7b": "application/x-pkcs7-certificates",
    ".p7c": "application/x-pkcs7-mime",
    ".p7m": "application/x-pkcs7-mime",
    ".p7r": "application/x-pkcs7-certreqresp",
    ".p7s": "application/x-pkcs7-signature",
    ".pbm": "image/x-portable-bitmap",
    ".pdb": "application/x-pilot",
    ".pdf": "application/pdf",
    ".pem": "application/x-x509-ca-cert",
    ".pfx": "application/x-pkcs12",
    ".pgm": "image/x-portable-graymap",
    ".pko": "application/ynd.ms-pkipko",
    ".pl": "application/x-perl",
    ".pm": "application/x-perl",
    ".pma": "application/x-perfmon",
    ".pmc": "application/x-perfmon",
    ".pml": "application/x-perfmon",
    ".pmr": "application/x-perfmon",
    ".pmw": "application/x-perfmon",
    ".png": "image/png",
    ".pnm": "image/x-portable-anymap",
    ".pot,": "application/vnd.ms-powerpoint",
    ".ppm": "image/x-portable-pixmap",
    ".pps": "application/vnd.ms-powerpoint",
    ".ppt": "application/vnd.ms-powerpoint",
    ".pptx": "application/vnd.openxmlformats-officedocument.presentationml.presentation",
    ".prc": "application/x-pilot",
    ".prf": "application/pics-rules",
    ".ps": "application/postscript",
    ".pub": "application/x-mspublisher",
    ".qt": "video/quicktime",
    ".ra": "audio/x-pn-realaudio",
    ".ram": "audio/x-pn-realaudio",
    ".rar": "application/x-rar-compressed",
    ".ras": "image/x-cmu-raster",
    ".rgb": "image/x-rgb",
    ".rmi": "audio/mid",
    ".roff": "application/x-troff",
    ".rpm": "application/x-redhat-package-manager",
    ".rss": "application/rss+xml",
    ".rtf": "application/rtf",
    ".rtx": "text/richtext",
    ".run": "application/x-makeself",
    ".scd": "application/x-msschedule",
    ".sct": "text/scriptlet",
    ".sea": "application/x-sea",
    ".setpay": "application/set-payment-initiation",
    ".setreg": "application/set-registration-initiation",
    ".sh": "application/x-sh",
    ".shar": "application/x-shar",
    ".shtml": "text/html",
    ".sit": "application/x-stuffit",
    ".snd": "audio/basic",
    ".spc": "application/x-pkcs7-certificates",
    ".spl": "application/futuresplash",
    ".src": "application/x-wais-source",
    ".sst": "application/vnd.ms-pkicertstore",
    ".stl": "application/vnd.ms-pkistl",
    ".stm": "text/html",
    ".sv4cpio": "application/x-sv4cpio",
    ".sv4crc": "application/x-sv4crc",
    ".svg": "image/svg+xml",
    ".svgz": "image/svg+xml",
    ".swf": "application/x-shockwave-flash",
    ".t": "application/x-troff",
    ".tar": "application/x-tar",
    ".tcl": "application/x-tcl",
    ".tex": "application/x-tex",
    ".texi": "application/x-texinfo",
    ".texinfo": "application/x-texinfo",
    ".tgz": "application/x-compressed",
    ".tif": "image/tiff",
    ".tiff": "image/tiff",
    ".tk": "application/x-tcl",
    ".tr": "application/x-troff",
    ".trm": "application/x-msterminal",
    ".ts": "video/mp2t",
    ".tsv": "text/tab-separated-values",
    ".txt": "text/plain",
    ".uls": "text/iuls",
    ".ustar": "application/x-ustar",
    ".vcf": "text/x-vcard",
    ".vrml": "x-world/x-vrml",
    ".war": "application/java-archive",
    ".wav": "audio/x-wav",
    ".wbmp": "image/vnd.wap.wbmp",
    ".wcm": "application/vnd.ms-works",
    ".wdb": "application/vnd.ms-works",
    ".webm": "video/webm",
    ".webp": "image/webp",
    ".wks": "application/vnd.ms-works",
    ".wmf": "application/x-msmetafile",
    ".wml": "text/vnd.wap.wml",
    ".wmlc": "application/vnd.wap.wmlc",
    ".wmv": "video/x-ms-wmv",
    ".woff": "application/font-woff",
    ".wps": "application/vnd.ms-works",
    ".wri": "application/x-mswrite",
    ".wrl": "x-world/x-vrml",
    ".wrz": "x-world/x-vrml",
    ".xaf": "x-world/x-vrml",
    ".xbm": "image/x-xbitmap",
    ".xhtml": "application/xhtml+xml",
    ".xla": "application/vnd.ms-excel",
    ".xlc": "application/vnd.ms-excel",
    ".xlm": "application/vnd.ms-excel",
    ".xls": "application/vnd.ms-excel",
    ".xlsx": "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet",
    ".xlt": "application/vnd.ms-excel",
    ".xlw": "application/vnd.ms-excel",
    ".xml": "text/xml",
    ".xof": "x-world/x-vrml",
    ".xpi": "application/x-xpinstall",
    ".xpm": "image/x-xpixmap",
    ".xspf": "application/xspf+xml",
    ".xwd": "image/x-xwindowdump",
    ".z": "application/x-compress",
    ".zip": "application/zip",
}


def urlsafe_b64encode(data):
    """
    urlsafe base64 encode

    :param data: string类型，待编码的字符串
    :return: string类型，编码的字符串
    """
    return s(base64.urlsafe_b64encode(b(data)))


def urlsafe_b64decode(data):
    """
    urlsafe base64 decode

    :param data: string类型，待解码的字符串
    :return: string类型，解码的字符串
    """
    return b(base64.urlsafe_b64decode(s(data)))


def standard_b64encode(data):
    """
    standard base64 encode

    :param data: string类型，待编码的字符串
    :return: string类型，编码的字符串
    """
    return s(base64.standard_b64encode(b(data)))


def standard_b64decode(data):
    """
    standard base64 decode

    :param data: string类型，待解码的字符串
    :return: string类型，解码的字符串
    """
    return b(base64.standard_b64decode(s(data)))


def _file_iter(input_stream, size):
    """
    二进制流迭代器

    :param input_stream: 二进制流
    :param size: integer类型，每次读取的块的大小
    :return: 指定大小的二进制块，如果读取失败会抛出IOError的异常
    """
    d = input_stream.read(size)
    while d:
        yield d
        d = input_stream.read(size)


def file_etag(localfile, size=BLOCKSIZE):
    """
    计算本地文件的etag

    :param localfile: string类型, 本地文件名
    :param size: integer类型, 分块大小
    :return: string类型, 本地文件的etag
    """

    filesize = os.path.getsize(localfile)
    blockcnt = filesize // size if filesize % size == 0 else filesize // size + 1

    hashstr = b''
    with open(localfile, 'rb') as input_stream:
        for block in _file_iter(input_stream, size):
            sha = hashlib.new('sha1')
            sha.update(b(block))
            hashstr += sha.digest()

    if blockcnt > 1:
        sha = hashlib.sha1()
        sha.update(hashstr)
        hashstr = sha.digest()

    return urlsafe_b64encode(struct.pack('@I', blockcnt) + hashstr)


def _check_dict(data):
    """
    check the type of data

    :param data: 键值对类型分别为string类型的dict类型
    :return: boolean类型，如果类型正确则返回True，否则抛出ValueError异常
    """
    if data is not None and isinstance(data, dict):
        return True
    raise ValueError('The input is not a dict-like object')


def ufile_put_url(bucket, key, upload_suffix=None):
    """
    采用普通上传方法上传UCloud UFile文件的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的文件名
    :return: string类型, 普通上传UFile的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)
    return 'http://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)


def ufile_post_url(bucket, upload_suffix=None):
    """
    采用表单上传方法上传UCloud UFile文件的url

    :param bucket: string类型, 待创建的空间名称
    :return: string类型, 表单上传UFile的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/'.format(bucket, upload_suffix or config.get_default('upload_suffix'))
    return 'http://{0}{1}/'.format(bucket, upload_suffix or config.get_default('upload_suffix'))


def ufile_uploadhit_url(bucket, upload_suffix=None):
    """
    秒传UCloud UFile文件的url

    :param bucket: string类型, 待创建的空间名称
    :return: string类型, 秒传UFile的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/uploadhit'.format(bucket, upload_suffix or config.get_default('upload_suffix'))
    return 'http://{0}{1}/uploadhit'.format(bucket, upload_suffix or config.get_default('upload_suffix'))


def initialsharding_url(bucket, key, upload_suffix=None):
    """
    初始化分片上传UCloud UFile的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的文件名
    :return: string类型, 初始化分片上传UFile的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}?uploads'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)
    return 'http://{0}{1}/{2}?uploads'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)


def finishsharding_url(bucket, key, upload_suffix=None):
    """
    结束分片上传UCloud UFile的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的文件名
    :return: string类型, 结束分片上传UFile的url
    """
    return ufile_put_url(bucket, key, upload_suffix=upload_suffix)


def shardingupload_url(bucket, key, uploadid, part_number, upload_suffix=None):
    """
    分片上传UCloud UFile的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的文件名
    :param uploadid: string类型, 初始化分片上传获得的uploadid字符串
    :param part_number: integer类型, 分片上传的编号,从0开始
    :return: string类型, 结束分片上传UFile的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}?uploadId={3}&partNumber={4}'.format(bucket,
                                                                upload_suffix or config.get_default('upload_suffix'),
                                                                key, uploadid, s(str(part_number)))
    return 'http://{0}{1}/{2}?uploadId={3}&partNumber={4}'.format(bucket,
                                                                  upload_suffix or config.get_default('upload_suffix'),
                                                                  key, uploadid, s(str(part_number)))


def ufile_getfilelist_url(bucket, upload_suffix=None):
    """
    获取文件列表的url

    :param bucket: string 类型，获取的空间名称
    :return: string类型，获取文件列表的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/?list'.format(bucket, upload_suffix or config.get_default('upload_suffix'))
    return 'http://{0}{1}/?list'.format(bucket, upload_suffix or config.get_default('upload_suffix'))


def mimetype_from_file(file):
    """
    获取文件的mimetype

    :param bucket: string 类型，获取的文件名称
    :return: string类型，获取文件的mimetype
    """
    ext = path.splitext(file)[1].lower()
    if ext in _EXTRA_TYPES_MAP:
        return _EXTRA_TYPES_MAP[ext]
    elif mimetypes.guess_type(file)[0] is not None:
        return mimetypes.guess_type(file)[0]

    return 'application/unknowntype'


def mimetype_from_buffer(stream):
    """
    获取流对象的mimetype

    :param bucket: string 类型，获取的流对象名称
    :return: string类型，获取流对象的mimetype
    """
    return 'application/octet-stream'


def ufile_restore_url(bucket, key, upload_suffix=None):
    """
    解冻冷存文件的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的文件名
    :return: string类型, 解冻文件的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}?restore'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)
    return 'http://{0}{1}/{2}?restore'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)


def ufile_classswitch_url(bucket, key, upload_suffix=None):
    """
    文件存储类型转换的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的文件名
    :return: string类型, 类型转换的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)
    return 'http://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)


def ufile_copy_url(bucket, key, upload_suffix=None):
    """
    拷贝文件的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的目标文件名
    :return: string类型, 拷贝文件的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)
    return 'http://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)


def ufile_rename_url(bucket, key, upload_suffix=None):
    """
    重命名文件的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的源文件名
    :param newkey:  string类型, 在空间中的目标文件名
    :return: string类型, 重命名文件的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)
    return 'http://{0}{1}/{2}'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)


def ufile_listobjects_url(bucket, upload_suffix=None):
    """
    获取目录文件列表的url

    :param bucket: string 类型，获取的空间名称
    :return: string类型，获取目录文件列表的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/?listobjects'.format(bucket, upload_suffix or config.get_default('upload_suffix'))
    return 'http://{0}{1}/?listobjects'.format(bucket, upload_suffix or config.get_default('upload_suffix'))


def ufile_listparts_url(bucket, upload_suffix, upload_id, max_parts=None, part_number_marker=None):
    """
    获取未完成分片上传的对象的已上传成功的分片列表。

    :param bucket: string类型, 空间名称
    :param upload_suffix: string类型, 域名后缀
    :param upload_id: string类型, 初始化分片上传获得的uploadid字符串
    :param max_parts: integer类型, 规定在US3响应中的最大Part数目。
    :param part_number_marker: integer类型, 指定List的起始位置，只有Part Number数目大于该参数的Part会被列出
    """
    if get_default("open_ssl"):
        url = 'https://{0}{1}/?muploadpart&uploadId={2}'.format(bucket,
                                                                upload_suffix or config.get_default('upload_suffix'),
                                                                upload_id)
    else:
        url = 'http://{0}{1}/?muploadpart&uploadId={2}'.format(bucket,
                                                               upload_suffix or config.get_default('upload_suffix'),
                                                               upload_id)
    if max_parts is not None:
        url += '&max-parts={}'.format(max_parts)
    if part_number_marker is not None:
        url += '&part-number-marker={}'.format(part_number_marker)
    return url


def ufile_op_meta_url(bucket, key, upload_suffix=None):
    """
    文件元数据的url

    :param bucket: string类型, 待创建的空间名称
    :param key:  string类型, 在空间中的文件名
    :return: string类型, 获取文件元数据的url
    """
    if get_default("open_ssl"):
        return 'https://{0}{1}/{2}?opmeta'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)
    return 'http://{0}{1}/{2}?opmeta'.format(bucket, upload_suffix or config.get_default('upload_suffix'), key)


def deprecated(message):
    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            warnings.warn("Call to deprecated function {} . -- {}".format(func.__name__, message),
                          category=DeprecationWarning,
                          stacklevel=2)
            warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator




# -*- coding: utf-8 -*-

import logging


LOG_FILE = 'ufile.log'
_formatter = logging.Formatter('[%(asctime)s, %(levelname)s]: %(filename)s-%(lineno)s: %(message)s')

_ch = logging.StreamHandler()
_ch.setLevel(logging.INFO)
_ch.setFormatter(_formatter)

logger = logging.getLogger("UCLOUD")
logger.setLevel(logging.INFO)
logger.addHandler(_ch)


def set_log_file(localfile=None):
    """
    设置日志的存放文件

    :param localfile: 存放日志文件名
    :return: None
    """
    global LOG_FILE
    if localfile is not None:
        LOG_FILE = localfile
    _fh = logging.FileHandler(LOG_FILE)
    _fh.setLevel(logging.INFO)
    _fh.setFormatter(_formatter)
    logger.addHandler(_fh)



# -*- coding: utf-8 -*-
import hashlib
import json
import os
import string
import time

from . import config
from .baseufile import BaseUFile
from .compact import b, s, u, url_parse, quote
from .config import BLOCKSIZE, get_default
from .httprequest import _put_stream, _put_file, _post_file, ResponseInfo, _uploadhit_file, _download_file, \
    _delete_file, _getfilelist, _head_file, _restore_file, _classswitch_file, _copy_file, _rename_file, _listobjects, \
    _op_meta
from .logger import logger
from .util import _check_dict, ufile_put_url, ufile_post_url, file_etag, ufile_uploadhit_url, ufile_getfilelist_url, \
    mimetype_from_file, ufile_restore_url, ufile_classswitch_url, ufile_copy_url, ufile_rename_url, \
    ufile_listobjects_url, ufile_op_meta_url


class FileManager(BaseUFile):
    """
    UCloud UFile普通上传文件类
    """

    def __init__(self, public_key, private_key, upload_suffix=None, download_suffix=None):
        """
        初始化 PutUFile 实例

        :param public_key: string类型, 账户API公私钥中的公钥
        :param private_key: string类型, 账户API公私钥中的私钥
        :param upload_suffix: string类型, 如果传入此参数, 则会忽略 config 中配置的 upload_suffix 字段
        :param download_suffix: string类型, 如果传入此参数, 则会忽略 config 中配置的 download_suffix 字段
        :return: None，如果为非法的公私钥，则抛出ValueError异常
        """
        super(FileManager, self).__init__(public_key, private_key)
        self.__upload_suffix = upload_suffix
        self.__download_suffix = download_suffix

    def _get_download_domain(self, bucket):
        if get_default("open_ssl"):
            return 'https://{0}{1}'.format(bucket, self.__download_suffix or config.get_default('download_suffix'))
        return 'http://{0}{1}'.format(bucket, self.__download_suffix or config.get_default('download_suffix'))

    def putstream(self, bucket, key, stream, mime_type=None, header=None):
        """
        上传二进制流到空间

        :param bucket: string类型，上传空间名称
        :param key:  string 类型，上传文件在空间中的名称
        :param stream: 二进制数据流,从文件指针位置开始发送数据,在调用时需调用者自己调整文件指针位置
        :param mime_type: 二进制数据流的MIME类型
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        if mime_type is None:
            mime_type = 'application/octet-stream'
        header['Content-Type'] = mime_type
        if config.get_default('md5') == True:
            m = hashlib.md5()
            m.update(stream.getvalue())
            header['Content-MD5'] = m.hexdigest()
            stream.seek(0, os.SEEK_SET)
        authorization = self.authorization('put', bucket, key, header)
        header['Authorization'] = authorization
        url = ufile_put_url(bucket, key, upload_suffix=self.__upload_suffix)
        logger.info('start put stream to bucket {0} as {1}'.format(bucket, key))
        logger.info('put UFile url: {0}'.format(url))
        logger.info('request header:\n{0}'.format(json.dumps(header, indent=4)))
        return _put_stream(url, header, stream)

    def putfile(self, bucket, key, localfile, header=None):
        """
        upload localfile to bucket as key

        :param bucket: string类型，上传空间名称
        :param key:  string 类型，上传文件在空间中的名称
        :param localfile: string类型，本地文件名称
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')
        mime_type = s(mimetype_from_file(localfile))
        file_size = os.path.getsize(localfile)
        header['Content-Type'] = mime_type
        if config.get_default('md5') == True:
            with open(localfile, 'rb') as data:
                m = hashlib.md5()
                m.update(data.read())
                header['Content-MD5'] = m.hexdigest()
                data.seek(0, os.SEEK_SET)
        authorization = self.authorization('put', bucket, key, header)
        header['Authorization'] = authorization
        if file_size != 0:
            header['Content-Length'] = str(file_size)
        url = ufile_put_url(bucket, key, upload_suffix=self.__upload_suffix)
        logger.info('start put file {0} to bucket {1} as {2}'.format(localfile, bucket, key))
        logger.info('put UFile url: {0}'.format(url))
        logger.info('request header:\n{0}'.format(json.dumps(header, indent=4)))
        return _put_file(url, header, localfile)

    def postfile(self, bucket, key, localfile, header=None):
        """
        表单上传文件到UFile空间

        :param bucket: string类型，上传空间名称
        :param key:  string 类型，上传文件在空间中的名称
        :param localfile: string类型，本地文件名称
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')
        mime_type = s(mimetype_from_file(localfile))

        # update the request header content-type
        boundary = self.__make_boundary()
        header['Content-Type'] = 'multipart/form-data; boundary={0}'.format(boundary)

        # form fields
        authorization = self.authorization('post', bucket, key, header, mime_type)
        fields = dict()
        fields['FileName'] = key
        fields['Authorization'] = authorization
        with open(localfile, 'rb') as stream:
            postdata = self.__make_postbody(boundary, fields, stream, mime_type, localfile)

        # update the request header content-length
        header['Content-Length'] = str(len(postdata))

        # post url
        url = ufile_post_url(bucket, upload_suffix=self.__upload_suffix)

        # start post file
        logger.info('start post file {0} to bucket {1} as {2}'.format(localfile, bucket, key))
        logger.info('post url is {0}'.format(url))

        return _post_file(url, header, postdata)

    def __make_boundary(self):
        """
        生成post内容主体的限定字符串

        :return:: string类型
        """
        t = time.time()
        m = hashlib.md5()
        m.update(b(str(t)))
        return m.hexdigest()

    def __make_postbody(self, boundary, fields, stream, mime_type, localfile):
        """
        生成post请求内容主体

        :param boundary: string类型，post内容主体的限定字符串
        :param fields: ditc类型，键值对类型分别为string类型
        :param stream: 可读的file-like object(file object 或者BytesIO)
        :param mime_type: string类型，上传文件或数据的MIME类型
        :param localfile: string类型，上传文件或数据的本地名称
        :return: 二进制数据流
        """

        binarystream = b''
        for (key, value) in fields.items():
            binarystream += b('--{0}\r\n'.format(boundary))
            binarystream += b('Content-Disposition: form-data; name="{0}"\r\n'.format(key))
            binarystream += b('\r\n')
            binarystream += b('{0}\r\n'.format(value))

        binarystream += b('--{0}\r\n'.format(boundary))
        binarystream += b('Content-Disposition: form-data; name="file"; filename="{0}"\r\n'.format(localfile))
        binarystream += b('Content-Type: {0}\r\n'.format(mime_type))
        binarystream += b('\r\n')

        binarystream += stream.read()
        binarystream += b('\r\n')
        binarystream += b('--{0}\r\n'.format(boundary))

        return binarystream

    def download_file(self, bucket, key, localfile, isprivate=True, expires=None, content_range=None, header=None):
        """
        下载UFile文件并且保存为本地文件

        :param bucket: string类型, UFile空间名称
        :param key: string类型， 下载文件在空间中的名称
        :param localfile: string类型，要保存的本地文件名称
        :param isprivate: boolean类型，如果为私有空间则为True
        :param expires: integer类型，私有文件链接有效时间
        :param content_range: tuple类型，元素为两个整型
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)

        if expires is None:
            expires = config.get_default('expires')

        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        if isinstance(content_range, tuple) and len(content_range) == 2:
            header['Range'] = 'bytes=' + '-'.join(map(lambda x: str(x), content_range))

        if not isprivate:
            url = self.public_download_url(bucket, key)
        else:
            url = self.private_download_url(bucket, key, expires, header, True)

        logger.info('get ufile url:{0}'.format(url))

        return _download_file(url, header, localfile)

    def public_download_url(self, bucket, key):
        """
        从公共空间下载文件的url

        :param bucket: string类型, 空间名称
        :param key: string类型，下载数据在空间中的名称
        :return: string类型，下载数据的url
        """
        return self._get_download_domain(bucket) + '/' + key

    def private_download_url(self, bucket, key, expires=None, header=None, internal=False):
        """
        从私有空间下载文件的url

        :param bucket: string类型, 空间名称
        :param key: string类型，下载数据在空间中的名称
        :param expires:  integer类型, 下载链接有效时间，单位为秒
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: string, 从私有空间下载文件和数据的url
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        if expires is None:
            expires = config.get_default('expires')
        expires += int(time.time())
        header['Expires'] = s(str(expires))

        signature = self.signature(bucket, key, 'get', header)
        query = {'UCloudPublicKey': self._public_key(),
                 'Expires': str(expires),
                 'Signature': signature}
        query_str = url_parse(query)

        url = self._get_download_domain(bucket) + '/' + key
        if internal:
            return url + '?' + query_str
        return url + '?UCloudPublicKey={0}&Expires={1}&Signature={2}'.format(self._public_key(), str(expires),
                                                                             signature)

    def private_head_url(self, bucket, key, expires=None, header=None):
        """
        从私有空间下载文件的url

        :param bucket: string类型, 空间名称
        :param key: string类型，下载数据在空间中的名称
        :param expires:  integer类型, 下载链接有效时间，单位为秒
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: string, 从私有空间下载文件和数据的url
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        if expires is None:
            expires = config.get_default('expires')
        expires += int(time.time())
        header['Expires'] = s(str(expires))

        signature = self.signature(bucket, key, 'head', header)
        return self._get_download_domain(bucket) + '/' + key + '?UCloudPublicKey={0}&Expires={1}&Signature={2}'.format(
            self._public_key(), str(expires), signature)

    def deletefile(self, bucket, key, header=None):
        """
        删除空间中文件方法

        :param bucket: string类型, 空间名称
        :param key:  string类型, 被删除文件在空间中的名称
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        authorization = self.authorization('delete', bucket, key, header)
        header['Authorization'] = authorization

        logger.info('start delete file {0} in bucket {1}'.format(key, bucket))
        url = ufile_put_url(bucket, key, upload_suffix=self.__upload_suffix)

        return _delete_file(url, header)

    def getfilelist(self, bucket, prefix=None, marker=None, limit=None, header=None):
        """
        获取bucket下的文件列表

        :param bucket: string 类型，空间名称
        :param prefix: string 类型，文件前缀, 默认为空字符串
        :param marker: string 类型，文件列表起始位置, 默认为空字符串
        :param limit: integer 类型，文件列表数目, 默认为20
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        header['Content-Length'] = str(0)
        authorization = self.authorization('get', bucket, '', header)
        header['Authorization'] = authorization
        param = dict()
        if marker is not None and isinstance(marker, str):
            param['marker'] = s(marker)
        if prefix is not None and isinstance(prefix, str):
            param['prefix'] = s(prefix)
        if limit is not None and isinstance(limit, int):
            param['limit'] = s(str(limit))
        info_message = ''.join(['start get file list from bucket {0}'.format(bucket),
                                '' if marker is None else ', marker: {0}'.format(
                                    marker if isinstance(marker, str) else marker.encode('utf-8')),
                                '' if limit is None else ', limit: {0}'.format(limit),
                                '' if prefix is None else ', prefix: {0}'.format(prefix)])
        logger.info(info_message)
        url = ufile_getfilelist_url(bucket, upload_suffix=self.__upload_suffix)
        return _getfilelist(url, header, param)

    def head_file(self, bucket, key, header=None):
        """
        获取空间中文件信息方法

        :param bucket: string类型, 空间名称
        :param key:  string类型, 文件在空间中的名称
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        authorization = self.authorization('head', bucket, key, header)
        header['Authorization'] = authorization

        logger.info('start head file {0} in bucket {1}'.format(key, bucket))
        url = ufile_put_url(bucket, key, upload_suffix=self.__upload_suffix)

        return _head_file(url, header)

    def compare_file_etag(self, bucket, remotekey, localfile):
        """
        比对空间文件和本地文件方法

        :param bucket: string类型, 空间名称
        :param remotekey:  string类型, 文件在空间中的名称
        :param localfile: string类型，本地文件的路径
        :return:True为比对一致，False为不一致
        """
        ret, resp = self.head_file(bucket, remotekey)
        remote_etag = resp.etag.strip('\"')
        local_etag = file_etag(localfile, BLOCKSIZE)
        return (remote_etag == local_etag)

    def restore_file(self, bucket, key, header=None):
        """
        解冻冷存文件方法

        :param bucket: string类型, 空间名称
        :param key:  string类型, 文件在空间中的名称
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        authorization = self.authorization('put', bucket, key, header)
        header['Authorization'] = authorization

        logger.info('start restore file {0} in bucket {1}'.format(key, bucket))
        url = ufile_restore_url(bucket, key, upload_suffix=self.__upload_suffix)

        return _restore_file(url, header)

    def class_switch_file(self, bucket, key, storageclass, header=None):
        """
        文件存储类型转换方法

        :param bucket: string类型, 空间名称
        :param key:  string类型, 文件在空间中的名称
        :param storageclass:  string类型
        尝试拷贝文件到UFile空间, 文件目标存储类型
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        authorization = self.authorization('put', bucket, key, header)
        header['Authorization'] = authorization

        # parameter
        params = {'storageClass': storageclass}

        logger.info('start switch file {0} storage class in bucket {1}'.format(key, bucket))
        url = ufile_classswitch_url(bucket, key, upload_suffix=self.__upload_suffix)

        return _classswitch_file(url, header, params)

    def copy(self, bucket, key, srcbucket, srckey, header=None):
        """
        尝试拷贝文件到UFile空间

        :param bucket: string类型，上传空间名称
        :param key:  string 类型，新文件在空间中的名称
        :param srcbucket: string类型，源文件所在空间名称
        :param srckey: string类型，源文件名称
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """

        if header is None:
            header = dict()
        _check_dict(header)
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        # update request header
        srckey = quote(srckey)
        header['X-Ufile-Copy-Source'] = "/" + srcbucket + "/" + srckey
        header['Content-Length'] = str(0)
        authorization = self.authorization('put', bucket, key, header)
        header['Authorization'] = authorization

        url = ufile_copy_url(bucket, key, upload_suffix=self.__upload_suffix)

        logger.info('start copy {0} in {1} to {2} in {3}'.format(srckey, srcbucket, key, bucket))
        logger.info('request url: {0}'.format(url))

        return _copy_file(url, header)

    def listobjects(self, bucket, prefix=None, marker=None, maxkeys=None, delimiter=None, header=None):
        """
        获取bucket下的文件列表

        :param bucket: string 类型，空间名称
        :param prefix: string 类型，返回以prefix作为前缀的目录文件列表
        :param marker: string 类型，返回以字母排序后，大于marker的目录文件列表
        :param maxkeys: integer 类型，指定返回目录文件列表的最大数量，默认值为100
        :param delimiter: stringr 类型，目录分隔符，当前只支持"/"和""，当delimiter设置为"/"时，返回目录形式的文件列表，当delimiter设置为""时，返回非目录层级文件列表
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if maxkeys is None:
            maxkeys = 100
        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')

        header['Content-Length'] = str(0)
        param = dict()
        if marker is not None and isinstance(marker, str):
            param['marker'] = s(marker)
        if prefix is not None and isinstance(prefix, str):
            param['prefix'] = s(prefix)
        if maxkeys is not None and isinstance(maxkeys, int):
            param['max-keys'] = s(str(maxkeys))
        if delimiter is not None and isinstance(delimiter, str):
            param['delimiter'] = s(delimiter)

        authorization = self.authorization('get', bucket, '', header, '', 'listobjects', param)
        header['Authorization'] = authorization

        info_message = ''.join(['start list objects from bucket {0}'.format(bucket),
                                '' if marker is None else ', marker: {0}'.format(
                                    marker if isinstance(marker, str) else marker.encode('utf-8')),
                                '' if maxkeys is None else ', maxkeys: {0}'.format(maxkeys),
                                '' if prefix is None else ', prefix: {0}'.format(prefix),
                                '' if delimiter is None else ', delimiter: {0}'.format(delimiter)])
        logger.info(info_message)
        url = ufile_listobjects_url(bucket, upload_suffix=self.__upload_suffix)
        return _listobjects(url, header, param)

    def opMeta(self, bucket, key, metak, metav, op="set", header=None):
        """
        获取文件元数据

        :param bucket: string 类型，空间名称
        :param key: string 类型，文件在空间中的名称
        :param metak: string 类型，元数据的key
        :param metav: string 类型，元数据的value
        :param op: string 类型，操作类型，目前只支持set
        :param header: dict类型，http 请求header，键值对类型分别为string，比如{'User-Agent': 'Google Chrome'}
        :return: ret: 如果http状态码为[200, 204, 206]之一则返回None，否则如果服务器返回json信息则返回dict类型，键值对类型分别为string, unicode string类型，否则返回空的dict
        :return:  ResponseInfo: 响应的具体信息，UCloud UFile 服务器返回信息或者网络链接异常
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)

        if 'User-Agent' not in header:
            header['User-Agent'] = config.get_default('user_agent')
        header['Content-Type'] = 'application/json'

        auth = self.authorization('post', bucket, key, header, action="opmeta")
        header['Authorization'] = auth
        data = {
            'op': op,
            'metak': metak,
            'metav': metav
        }
        url = ufile_op_meta_url(bucket, key, upload_suffix=self.__upload_suffix)
        return _op_meta(url, json.dumps(data), header)



# -*- coding: utf-8 -*-

import string
from .auth import Auth
from .util import _check_dict


class BaseUFile(object):
    """
    UCloud UFile 内容管理的基类，主要包括不同操作的公共方法
    """
    def __init__(self, public_key, private_key):
        """
        初始化 BaseUFile 对象

        :param public_key: string类型， 账户API公私钥的公钥
        :param private_key: string类型，账户API公私钥的私钥
        :return: None，如果为非法的公私钥则抛出ValueError异常
        """
        self.__auth = Auth(public_key, private_key)

    def set_keys(self, public_key, private_key):
        """
        重新设置账户API公私钥

        :param public_key: string类型， 账户API公私钥的公钥
        :param private_key: string类型，账户API公私钥的私钥
        :return: None，如果为非法的公私钥则抛出ValueError异常
        """
        self.__auth.set_keys(public_key, private_key)

    def authorization(self, method, bucket, key, header=None, mime_type=None, action=None, url_param=None):
        """
        根据不同的上传方法和HTTP请求头获得上传凭证

        :param method: string类型，文件上传方法，为'put'或'post'
        :param bucket: string类型，上传的空间名称
        :param key: string类型，文件在上传空间中的名称
        :param header: dict类型，键值对类型分别为string类型，HTTP请求的header
        :param mime_type: string类型，上传文件的MIME，如果提供则采用其计算文件上传凭证
        :param action: string类型，请求动作
        :param url_param: dict类型，键值对分别为string类型，HTTP请求url参数
        :return: string类型，本次文件上传的上传凭证
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        if action is None:
            action = ''
        if url_param is None:
            url_param = dict()
        data = self.__digest_authorization_data(method, bucket, key, header, mime_type, action, url_param)
        return self.__auth.ufile_authorization(data)

    def signature(self, bucket, key, method="get", header=None, mime_type=None):
        """
        根据不同的HTTP请求headers获得文件下载凭证

        :param bucket: string类型, 下载文件所在空间名称
        :param key:    string类型，现在文件在空间中名称
        :param method: string类型，为'get'
        :param header: dict类型，键值对类型分别为string类型，HTTP请求的header
        :param mime_type: string类型，上传文件的MIME，如果提供则采用其计算文件上传凭证
        :return: string类型，本次文件下载的下载签名
        """

        if header is None:
            header = dict()
        else:
            _check_dict(header)
        data = self.__digest_signature_data(bucket, key, method, header, mime_type)
        return self.__auth.ufile_signature(data)

    def __digest_authorization_data(self, method, bucket, key, header=None, mime_type=None, action=None, url_param=None):
        """
        获得进行认证的字符串

        :param method: string类型, 为'put','post'和 'delete'
        :param bucket: string类型，空间名称
        :param key:    string类型，文件在空间中的名称
        :param header: dict类型，键值对类型分别为string类型，HTTP请求的header
        :param mime_type: string类型，上传文件的MIME，如果提供则采用其计算文件上传凭证
        :param action: string类型，请求动作
        :param url_param: dict类型，键值对分别为string类型，HTTP请求url参数
        :return: string类型，将要进行认证的字符串
        """
        if header is None:
            header = dict()
        else:
            _check_dict(header)
        data = ''.join([method.upper(), '\n'])
        data += ''.join(['' if 'Content-MD5' not in header else header['Content-MD5'], '\n'])
        data += ''.join([mime_type if mime_type is not None else '' if 'Content-Type' not in header else header['Content-Type'], '\n'])
        data += ''.join(['' if 'Date' not in header else header['Date'], '\n'])
        data += ''.join([self.__canonicalize_ucloud_headers(header), self.__canonicalize_resource(bucket, key), self.__canonicalize_url_params(action, url_param)])
        return data


    def __digest_signature_data(self, bucket, key, method='get', header=None, mime_type=None):
        """
        获得进行下载签名的字符串

        :param bucket: string类型， 空间名称
        :param key: string类型， 文件在空间中的名称
        :param method: string类型，'get'
        :param header: dict类型，键值对类型分别为string类型，HTTP请求的header
        :param mime_type: string类型，上传文件的MIME，如果提供则采用其计算文件上传凭证
        :return: string类型，将要进行签名的字符串
        """
        data = ''.join([method.upper(), '\n'])
        data += ''.join(['' if 'Content-MD5' not in header else header['Content-MD5'], '\n'])
        data += ''.join([mime_type if mime_type is not None else '' if 'Content-Type' not in header else header['Content-Type'], '\n'])
        data += ''.join(['' if 'Expires' not in header else header['Expires'], '\n'])
        data += ''.join([self.__canonicalize_ucloud_headers(header), self.__canonicalize_resource(bucket, key)])
        return data

    def __canonicalize_ucloud_headers(self, header):
        """
        字符串化UCloud附加标头

        :param header: dict类型，键值对类型分别为string类型，UCloud附加标头，键名以'X-UCloud-'开头
        :return: string类型
        """
        # 以下语法不适用于Python 2.6
        # ucloud_headers_map = {string.lower(x).strip(): x for x in header if string.lower(x).strip().startswith('x-ucloud-')}
        ucloud_headers_map = dict([(x.lowerx.strip(), x) for x in header if x.lower().strip().startswith('x-ucloud-')])
        ucloud_keys = sorted(ucloud_headers_map)
        return '\n'.join([x+':'+header[ucloud_headers_map[x]].strip() for x in ucloud_keys])

    def __canonicalize_url_params(self, action, url_param):
        """
        字符串化url中的参数

        :param action: string类型，请求动作
        :param url_param: dict类型，键值对分别为string类型，HTTP请求url参数
        :return: string类型
        """
        if action == 'listobjects':
            return ''.join('\nlistobjects\ndelimiter:{0}\nmarker:{1}\nmax-keys:{2}\nprefix:{3}'.format(url_param['delimiter'], url_param['marker'], url_param['max-keys'],  url_param['prefix']))
        return ''

    def __canonicalize_resource(self, bucket, key):
        """
        获得UFile 资源字符串

        :param bucket: string类型, 空间名称
        :param key: string类型，文件在空间中的名称
        :return: string类型，UFile资源字符串
        """
        return '/{0}/{1}'.format(bucket, key)

    def _public_key(self):
        """
        返回账户公私钥中的公钥

        :return: string类型，公钥
        """
        return self.__auth._public_key()
